@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/child.c,v 1.15.2.1 2004/04/08 14:14:30 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/child.c,v 1.21 2004/04/21 00:11:34 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -52,7 +52,10 @@ static void unset_nonblock(int fd);
 static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd);
 static StartupPacket *read_startup_packet(POOL_CONNECTION *cp);
 static int send_startup_packet(POOL_CONNECTION *cp, StartupPacket *sp);
+static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION *frontend);
 static void reset_backend(POOL_CONNECTION *backend);
+static void cancel_request(CancelPacket *cp, int secondary_backend);
+static RETSIGTYPE die(int sig);
 
 /*
 * child main loop
@@ -64,13 +67,13 @@ void do_child(int unix_fd, int inet_fd)
 	POOL_CONNECTION_POOL *backend;
 
 	pool_debug("I am %d", getpid());
-
 	/* set up signal handlers */
 	signal(SIGALRM, SIG_DFL);
-	signal(SIGTERM, SIG_DFL);
+	signal(SIGTERM, die);
 	signal(SIGINT, SIG_DFL);
 	signal(SIGCHLD, SIG_DFL);
 	signal(SIGUSR1, SIG_DFL);
+	signal(SIGPIPE, SIG_IGN);
 
 #ifdef NONE_BLOCK
 	/* set listen fds to none block */
@@ -101,9 +104,12 @@ void do_child(int unix_fd, int inet_fd)
 			continue;
 		}
 
-		/* unset frontend fd tp none block */
+		/* set frontend fd to blocking */
 		unset_nonblock(frontend->fd);
 
+		/* disable timeout */
+		pool_disable_timeout();
+
 		/* read the startup packet */
 	retry_startup:
 		sp = read_startup_packet(frontend);
@@ -138,34 +144,9 @@ void do_child(int unix_fd, int inet_fd)
 		/* cancel request? */
 		if (ntohl(sp->protoVersion) == 0x04D2162E)		/* 0x04D2162F = 12345678 */
 		{
-			int	len;
-			int fd;
-			POOL_CONNECTION *con;
-			CancelPacket *cp = (CancelPacket *)sp;
-
-			pool_debug("Cancel request received");
-
-			if (*pool_config.current_backend_host_name == '\0')
-				fd = connect_unix_domain_socket((POOL_CONNECTION_POOL *)NULL);
-			else
-				fd = connect_inet_domain_socket((POOL_CONNECTION_POOL *)NULL);
-
-			if (fd < 0)
-			{
-				pool_error("Could not create socket for sending cancel request");
-				continue;
-			}
-
-			con = pool_open(fd);
-			if (con == NULL)
-				continue;
-
-			len = htonl(sizeof(len) + sizeof(CancelPacket));
-			pool_write(con, &len, sizeof(len));
-
-			if (pool_write_and_flush(con, cp, sizeof(CancelPacket)) < 0)
-				pool_error("Could not send cancel request packet");
-			pool_close(con);
+			cancel_request((CancelPacket *)sp, 0);
+			if (pool_config.replication_enabled)
+				cancel_request((CancelPacket *)sp, 1);
 			continue;
 		}
 
@@ -197,47 +178,17 @@ void do_child(int unix_fd, int inet_fd)
 		{
 			connection_reuse = 0;
 
-			/* connect to the backend and send the startup packect here */
-			backend = pool_create_cp(sp->user, sp->database);
-			if (backend == NULL)
-			{
-				static char *msg = "Sorry, too many clients already";
-
-				pool_error("do_child: cannot not create backend connection");
-
-				pool_write(frontend, "E", 1);
-				pool_write_and_flush(frontend, msg, strlen(msg)+1);
-				pool_close(frontend);
+			if ((backend = connect_backend(sp, frontend)) == NULL)
 				continue;
-			}
-
-			/* mark this is a backend connection */
-			backend->con->isbackend = 1;
-
-			/* send startup packet */
-			if (send_startup_packet(backend->con, sp) < 0)
-			{
-				pool_error("do_child: fails to send startup packet to the backend");
-				pool_close(frontend);
-				continue;
-			}
-
-			/*
-			 * do authentication stuff
-			 */
-			if (pool_do_auth(frontend, backend))
-			{
-				pool_close(frontend);
-				pool_discard_cp(sp->user, sp->database);
-				continue;
-			}
 		}
 		else
 		{
-			reset_backend(backend->con);
+			reset_backend(MASTER_CONNECTION(backend)->con);
+			if (pool_config.replication_enabled)
+				reset_backend(SECONDARY_CONNECTION(backend)->con);
 
 			/* send Auth OK and Ready for Query to the frontend */
-			if (pool_send_auth_ok(frontend, backend->pid, backend->key) != POOL_CONTINUE)
+			if (pool_send_auth_ok(frontend, MASTER_CONNECTION(backend)->pid, MASTER_CONNECTION(backend)->key) != POOL_CONTINUE)
 			{
 				pool_close(frontend);
 				continue;
@@ -249,36 +200,44 @@ void do_child(int unix_fd, int inet_fd)
 			}
 		}
 
+		/* enable query result read timeout if non strict mode */
+		if (pool_config.replication_strict == 0)
+			pool_enable_timeout();
+
 		/* query process loop */
 		for (;;)
 		{
 			int status;
 
-			status = pool_process_query(frontend, backend->con, 
-										connection_reuse);
+			status = pool_process_query(frontend, backend, connection_reuse);
 			switch (status)
 			{
 				/* client exits */
 				case POOL_END:
+					pool_close(frontend);
+
 					/* do not cache connection to template0, template1, regression */
 					if (!strcmp(sp->database, "template0") || !strcmp(sp->database, "template1") ||
 						!strcmp(sp->database, "regression"))
 						pool_discard_cp(sp->user, sp->database);
-
-					pool_close(frontend);
-
-					pool_connection_pool_timer(backend);
+					else
+						pool_connection_pool_timer(backend);
 					break;
 				
 				/* error occured. discard backend connection pool
                    and disconnect connection to the frontend */
 				case POOL_ERROR:
+#ifdef NOT_USED
 					pool_discard_cp(sp->user, sp->database);
 					pool_close(frontend);
+#endif
+					notice_backend_error();
+					exit(1);
 					break;
 
 				/* fatal error occured. just exit myself... */
 				case POOL_FATAL:
+					notice_backend_error();
 					exit(1);
 					break;
 
@@ -352,7 +311,6 @@ static void unset_nonblock(int fd)
 static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd)
 {
     fd_set	readmask;
-    fd_set	exceptmask;
     int fds;
 
 	struct sockaddr addr;
@@ -363,20 +321,15 @@ static POOL_CONNECTION *do_accept(int unix_fd, int inet_fd)
 	POOL_CONNECTION *cp;
 
 	FD_ZERO(&readmask);
-	FD_ZERO(&exceptmask);
 	FD_SET(unix_fd, &readmask);
 	if (inet_fd)
 		FD_SET(inet_fd, &readmask);
-	FD_SET(unix_fd, &exceptmask);
-	if (inet_fd)
-		FD_SET(inet_fd, &exceptmask);
 
-	fds = select(Max(unix_fd, inet_fd)+1,
-				 &readmask, NULL, &exceptmask, NULL);
+	fds = select(Max(unix_fd, inet_fd)+1, &readmask, NULL, NULL, NULL);
 
 	if (fds == -1)
 	{
-		if (errno == EINTR)
+		if (errno == EAGAIN || errno == EINTR)
 			return NULL;
 
 		pool_error("select() failed. reason %s", strerror(errno));
@@ -533,3 +486,106 @@ static void reset_backend(POOL_CONNECTION *backend)
 		}
 	}
 }
+
+static void cancel_request(CancelPacket *cp, int secondary_backend)
+{
+	int	len;
+	int fd;
+	POOL_CONNECTION *con;
+
+	pool_debug("Cancel request received");
+
+	if (secondary_backend)
+	{
+		if (*pool_config.secondary_backend_host_name == '\0')
+			fd = connect_unix_domain_socket(1);
+		else
+			fd = connect_inet_domain_socket(1);
+	}
+	else
+	{
+		if (*pool_config.current_backend_host_name == '\0')
+			fd = connect_unix_domain_socket(0);
+		else
+			fd = connect_inet_domain_socket(0);
+	}
+
+	if (fd < 0)
+	{
+		pool_error("Could not create socket for sending cancel request");
+		return;
+	}
+
+	con = pool_open(fd);
+	if (con == NULL)
+		return;
+
+	len = htonl(sizeof(len) + sizeof(CancelPacket));
+	pool_write(con, &len, sizeof(len));
+
+	if (pool_write_and_flush(con, cp, sizeof(CancelPacket)) < 0)
+		pool_error("Could not send cancel request packet");
+	pool_close(con);
+}
+
+static POOL_CONNECTION_POOL *connect_backend(StartupPacket *sp, POOL_CONNECTION *frontend)
+{
+	POOL_CONNECTION_POOL *backend;
+
+	/* connect to the backend and send the startup packect here */
+	backend = pool_create_cp(sp->user, sp->database);
+	if (backend == NULL)
+	{
+		static char *msg = "Sorry, too many clients already";
+
+		pool_error("do_child: cannot not create backend connection");
+
+		pool_write(frontend, "E", 1);
+		pool_write_and_flush(frontend, msg, strlen(msg)+1);
+		pool_close(frontend);
+		return NULL;
+	}
+
+	/* mark this is a backend connection */
+	backend->slots[0]->con->isbackend = 1;
+	if (pool_config.replication_enabled)
+	{
+		backend->slots[1]->con->isbackend = 1;
+		backend->slots[1]->con->issecondary_backend = 1;
+	}
+
+	/* send startup packet */
+	if (send_startup_packet(backend->slots[0]->con, sp) < 0)
+	{
+		pool_error("do_child: fails to send startup packet to the backend");
+		pool_close(frontend);
+		return NULL;
+	}
+
+	/* send startup packet */
+	if (pool_config.replication_enabled)
+	{
+		if (send_startup_packet(backend->slots[1]->con, sp) < 0)
+		{
+			pool_error("do_child: fails to send startup packet to the secondary backend");
+			pool_close(frontend);
+			return NULL;
+		}
+	}
+
+	/*
+	 * do authentication stuff
+	 */
+	if (pool_do_auth(frontend, backend))
+	{
+		pool_close(frontend);
+		pool_discard_cp(sp->user, sp->database);
+		return NULL;
+	}
+	return backend;
+}
+
+static RETSIGTYPE die(int sig)
+{
+	exit(0);
+}
