@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/main.c,v 1.6 2004/03/24 00:07:01 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/main.c,v 1.9 2004/04/21 00:11:34 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -45,6 +45,7 @@
 
 #define PGPOOLMAXLITSENQUEUELENGTH 10000
 static void daemonize(void);
+static int read_pid_file(void);
 static void write_pid_file(void);
 static pid_t fork_a_child(int unix_fd, int inet_fd);
 static int create_unix_domain_socket(void);
@@ -84,6 +85,7 @@ int main(int argc, char **argv)
 	int opt;
 	char conf_file[POOLMAXPATHLEN+1];
 	int i;
+	int pid;
 
 	snprintf(conf_file, sizeof(conf_file), "%s/%s", DEFAULT_CONFIGDIR, POOL_CONF_FILE_NAME);
 
@@ -133,14 +135,16 @@ int main(int argc, char **argv)
 	}
 	else if (optind == argc)
 	{
-		char path[POOLMAXPATHLEN];
-
-		snprintf(path, sizeof(path), "%s/%s", pool_config.logdir, PID_FILE_NAME);
-
-		if (check_pid_file(path) == 0)
+		pid = read_pid_file();
+		if (pid > 0)
 		{
-			fprintf(stderr, "pid file %s found. is another pgpool is running?\n", path);
-			exit(1);
+			if (kill(pid, 0) == 0)
+			{
+				fprintf(stderr, "pid file found. is another pgpool(%d) is running?\n", pid);
+				exit(1);
+			}
+			else
+				fprintf(stderr, "pid file found but it seems bogus. Trying to start pgpool anyway...\n");
 		}
 	}
 	else if (optind < argc)
@@ -203,7 +207,7 @@ int main(int argc, char **argv)
 
 static void usage(void)
 {
-	fprintf(stderr, "pgpool version %s, a generic connection pool server for PostgreSQL\n\n", PGPOOLVERSION);
+	fprintf(stderr, "pgpool version %s, a generic connection pool/replication server for PostgreSQL\n\n", PGPOOLVERSION);
 	fprintf(stderr, "usage: pgpool [-f config_file][-n][-d][-h][stop]\n");
 	fprintf(stderr, "  config_file default path: %s/%s\n",DEFAULT_CONFIGDIR, POOL_CONF_FILE_NAME);
 	fprintf(stderr, "  -n: don't run in daemon mode. does not detatch control tty\n");
@@ -290,10 +294,36 @@ static int check_pid_file(char *path)
 	return stat(path, &buf);
 }
 
+/*
+* read the pid file
+*/
+static int read_pid_file(void)
+{
+	FILE *fd;
+	char path[POOLMAXPATHLEN];
+	char pidbuf[128];
+
+	snprintf(path, sizeof(path), "%s/%s", pool_config.logdir, PID_FILE_NAME);
+	fd = fopen(path, "r");
+	if (!fd)
+	{
+		return -1;
+	}
+	if (fread(pidbuf, 1, sizeof(pidbuf), fd) <= 0)
+	{
+		pool_error("could not read pid file as %s. reason: %s",
+				   path, strerror(errno));
+		fclose(fd);
+		return -1;
+	}
+	fclose(fd);
+	return(atoi(pidbuf));
+}
+
 /*
 * write the pid file
 */
-static void write_pid_file()
+static void write_pid_file(void)
 {
 	FILE *fd;
 	char path[POOLMAXPATHLEN];
@@ -445,6 +475,8 @@ void notice_backend_error(void)
 	pid_t parent = getppid();
 
 	kill(parent, SIGUSR1);
+	/* avoid race conditon with SIGCHLD */
+	sleep(1);
 }
 
 static RETSIGTYPE exit_handler(int sig)
@@ -475,6 +507,7 @@ static RETSIGTYPE exit_handler(int sig)
 static RETSIGTYPE failover_handler(int sig)
 {
 	int i;
+	int replication = 0;
 
 	pool_debug("failover_handler called");
 
@@ -487,6 +520,11 @@ static RETSIGTYPE failover_handler(int sig)
 	/* secondary backend exists? */
 	if (pool_config.secondary_backend_port == 0)
 		return;
+#ifdef NOT_USED
+	/* in replication mode and already degeneration done */
+	else if (pool_config.replication_mode && !pool_config.replication_enabled)
+		return;
+#endif
 
 	/* already failover */
 	if (strcmp(pool_config.current_backend_host_name, pool_config.secondary_backend_host_name) ||
@@ -494,11 +532,21 @@ static RETSIGTYPE failover_handler(int sig)
 	{
 		switching = 1;
 		
-		pool_debug("starting failover from %s(%d) to %s(%d)",
-				   pool_config.current_backend_host_name,
-				   pool_config.current_backend_port,
-				   pool_config.secondary_backend_host_name,
-				   pool_config.secondary_backend_port);
+		if (pool_config.replication_enabled)
+		{
+			replication = 1;
+			pool_log("starting degenration. shutdown secondary host %s(%d)",
+					   pool_config.secondary_backend_host_name,
+					   pool_config.secondary_backend_port);
+		}
+		else
+		{
+			pool_log("starting failover from %s(%d) to %s(%d)",
+					   pool_config.current_backend_host_name,
+					   pool_config.current_backend_port,
+					   pool_config.secondary_backend_host_name,
+					   pool_config.secondary_backend_port);
+		}
 
 		/* kill all children */
 		for (i = 0; i < pool_config.num_init_children; i++)
@@ -513,9 +561,17 @@ static RETSIGTYPE failover_handler(int sig)
 		if (errno != ECHILD)
 			pool_error("wait() failed. reason:%s", strerror(errno));
 
-		/* fail over to secondary */
-		pool_config.current_backend_host_name = pool_config.secondary_backend_host_name;
-		pool_config.current_backend_port = pool_config.secondary_backend_port;
+		if (pool_config.replication_enabled)
+		{
+			/* disable replicaton mode */
+			pool_config.replication_enabled = 0;
+		}
+		else
+		{
+			/* fail over to secondary */
+			pool_config.current_backend_host_name = pool_config.secondary_backend_host_name;
+			pool_config.current_backend_port = pool_config.secondary_backend_port;
+		}
 
 		/* fork the children */
 		for (i=0;i<pool_config.num_init_children;i++)
@@ -525,11 +581,20 @@ static RETSIGTYPE failover_handler(int sig)
 
 		switching = 0;
 
-		pool_debug("failover from %s(%d) to %s(%d) done.",
-				   pool_config.backend_host_name,
-				   pool_config.backend_port,
-				   pool_config.secondary_backend_host_name,
-				   pool_config.secondary_backend_port);
+		if (replication)
+		{
+			pool_log("degenration done. shutdown secondary host %s(%d)",
+					   pool_config.secondary_backend_host_name,
+					   pool_config.secondary_backend_port);
+		}
+		else
+		{
+			pool_log("failover from %s(%d) to %s(%d) done.",
+					   pool_config.backend_host_name,
+					   pool_config.backend_port,
+					   pool_config.secondary_backend_host_name,
+					   pool_config.secondary_backend_port);
+		}
 	}
 }
 
@@ -562,8 +627,11 @@ static RETSIGTYPE reap_handler(int sig)
 				if (pid == pids[i])
 				{
 					/* if found, fork a new child */
-					pids[i] = fork_a_child(unix_fd, inet_fd);
-					break;
+					if (!switching && !exiting)
+					{
+						pids[i] = fork_a_child(unix_fd, inet_fd);
+						break;
+					}
 				}
 			}
 		}
