@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/pool_auth.c,v 1.2 2004/03/24 00:07:01 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/pool_auth.c,v 1.4 2004/03/31 07:57:31 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -42,54 +42,99 @@
 POOL_STATUS pool_do_auth(POOL_CONNECTION *frontend,
 						 POOL_CONNECTION_POOL *cp)
 {
-	POOL_CONNECTION *backend = cp->con;
 	int status;
-	char kind;
-	int pid;
-	int key;
+	char kind, kind1;
+	int pid, pid1;
+	int key, key1;
 
-	status = pool_read(backend, &kind, 1);
+	status = pool_read(MASTER_CONNECTION(cp)->con, &kind, 1);
 	if (status < 0)
 	{
 		pool_error("pool_do_auth: error while reading \"R\"", strerror(errno));
 		return -1;
 	}
-	/* Error response? */
-	else if (kind == 'E')
+
+	if (pool_config.replication_enabled)
 	{
-		ErrorResponse(frontend, backend);
+		status = pool_read(SECONDARY_CONNECTION(cp)->con, &kind1, 1);
+		if (status < 0)
+		{
+			pool_error("pool_do_auth: error while reading \"R\" from secondary backend",
+					   strerror(errno));
+			return -1;
+		}
+
+		if (kind != kind1)
+		{
+			pool_error("pool_do_auth: kind is not match between backends master(%c) secondary(%c)",
+					   kind, kind1);
+			return -1;
+		}
+	}
+
+	/* Error response? */
+	if (kind == 'E')
+	{
+		ErrorResponse(frontend, cp);
 		return -1;
 	}
-	if (kind != 'R')
+	else if (kind != 'R')
 	{
 		pool_error("pool_do_auth: expect \"R\" got %c", kind);
 		return -1;
 	}
 
-	status = pool_read(backend, &pid, sizeof(pid));
+	status = pool_read(MASTER_CONNECTION(cp)->con, &pid, sizeof(pid));
 
 	if (status < 0 || pid != 0)
 	{
 		pool_error("pool_do_auth: backend does not return Authenticaton Ok");
 		return -1;
 	}
+	if (pool_config.replication_enabled)
+	{
+		status = pool_read(SECONDARY_CONNECTION(cp)->con, &pid1, sizeof(pid1));
+
+		if (status < 0 || pid1 != 0)
+		{
+			pool_error("pool_do_auth: secondary backend does not return Authenticaton Ok");
+			return -1;
+		}
+	}
 
 	/*
 	 * Authentication OK. Now read pid and secret key from the
 	 * backend
 	 */
 	kind = 'W';
-	status = pool_read(backend, &kind, 1);
+	status = pool_read(MASTER_CONNECTION(cp)->con, &kind, 1);
 	if (status < 0)
 	{
 		pool_error("pool_do_auth: error while reading \"W\"");
 		return -1;
 	}
+	if (pool_config.replication_enabled)
+	{
+		status = pool_read(SECONDARY_CONNECTION(cp)->con, &kind1, 1);
+		if (status < 0)
+		{
+			pool_error("pool_do_auth: error while reading \"W\" from secondary backend",
+					   strerror(errno));
+			return -1;
+		}
+
+		if (kind != kind1)
+		{
+			pool_error("pool_do_auth: kind is not match between backends master(%c) secondary(%c)",
+					   kind, kind1);
+			return -1;
+		}
+	}
 
 	/* Error response? */
 	if (kind == 'E')
 	{
-		ErrorResponse(frontend, backend);
+		ErrorResponse(frontend, cp);
 		return -1;
 	}
 	else if (kind != 'K')
@@ -102,12 +147,22 @@ POOL_STATUS pool_do_auth(POOL_CONNECTION *frontend,
 	 * OK, read pid and secret key
 	 */
 	/* pid */
-	pool_read(backend, &pid, sizeof(pid));
-	cp->pid = pid;
+	pool_read(MASTER_CONNECTION(cp)->con, &pid, sizeof(pid));
+	MASTER_CONNECTION(cp)->pid = pid;
 
 	/* key */
-	pool_read(backend, &key, sizeof(key));
-	cp->key = key;
+	pool_read(MASTER_CONNECTION(cp)->con, &key, sizeof(key));
+	MASTER_CONNECTION(cp)->key = key;
+
+	if (pool_config.replication_enabled)
+	{
+		pool_read(SECONDARY_CONNECTION(cp)->con, &pid1, sizeof(pid1));
+		SECONDARY_CONNECTION(cp)->pid = pid;
+
+		/* key */
+		pool_read(SECONDARY_CONNECTION(cp)->con, &key1, sizeof(key1));
+		SECONDARY_CONNECTION(cp)->key = key;
+	}
 
 	return (pool_send_auth_ok(frontend, pid, key));
 }
