@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
- * $Header: /home/t-ishii/repository/pgpool/pool_connection_pool.c,v 1.10 2004/03/24 02:26:51 t-ishii Exp $
+ * $Header: /home/t-ishii/repository/pgpool/pool_connection_pool.c,v 1.15 2004/04/14 00:28:28 t-ishii Exp $
  *
  * pgpool: a language independent connection pool server for PostgreSQL 
  * written by Tatsuo Ishii
@@ -41,6 +41,9 @@
 
 POOL_CONNECTION_POOL *pool_connection_pool;	/* connection pool */
 
+static POOL_CONNECTION_POOL_SLOT *create_cp(POOL_CONNECTION_POOL_SLOT *cp,
+											char *user, char *database,
+											int secondary_backend);
 /*
 * initialize connection pool
 */
@@ -74,12 +77,12 @@ POOL_CONNECTION_POOL *pool_get_cp(char *user, char *database)
 
 	for (i=0;i<pool_config.max_pool;i++)
 	{
-		if (p->user != NULL &&
-			strcmp(p->user, user) == 0 &&
-			strcmp(p->database, database) == 0)
+		if (MASTER_CONNECTION(p) && MASTER_CONNECTION(p)->user != NULL &&
+			strcmp(MASTER_CONNECTION(p)->user, user) == 0 &&
+			strcmp(MASTER_CONNECTION(p)->database, database) == 0)
 		{
 			/* mark this connection is under use */
-			p->closetime = 0;
+			MASTER_CONNECTION(p)->closetime = 0;
 			return p;
 		}
 		p++;
@@ -102,12 +105,21 @@ void pool_discard_cp(char *user, char *database)
 		return;
 	}
 
-	free(p->user);
-	free(p->database);
-	pool_close(p->con);
+	free(MASTER_CONNECTION(p)->user);
+	free(MASTER_CONNECTION(p)->database);
+	pool_close(MASTER_CONNECTION(p)->con);
+
+	if (pool_config.replication_enabled)
+	{
+		free(SECONDARY_CONNECTION(p)->user);
+		free(SECONDARY_CONNECTION(p)->database);
+		pool_close(SECONDARY_CONNECTION(p)->con);
+	}
+
 	memset(p, 0, sizeof(POOL_CONNECTION_POOL));
 }
 
+
 /*
 * create connection pool by user and database
 */
@@ -120,43 +132,27 @@ POOL_CONNECTION_POOL *pool_create_cp(char *user, char *database)
 
 	if (p == NULL)
 	{
-		pool_error("pool_init_cp: malloc() failed");
+		pool_error("pool_create_cp: pool_connection_pool is not initialized");
 		return NULL;
 	}
 
 	for (i=0;i<pool_config.max_pool;i++)
 	{
-		if (p->user == NULL)
+		if (MASTER_CONNECTION(p) == NULL)
 		{
-			pool_debug(":%s:", pool_config.current_backend_host_name);
+			/* create master connection */
+			MASTER_CONNECTION(p) = malloc(sizeof(POOL_CONNECTION_POOL_SLOT));
+			MASTER_CONNECTION(p) = create_cp(MASTER_CONNECTION(p), user, database, 0);
+			p->num = 1;
 
-			if (*pool_config.current_backend_host_name == '\0')
-				fd = connect_unix_domain_socket(p);
-			else
-				fd = connect_inet_domain_socket(p);
-
-			if (fd < 0)
+			/* create secondary connection */
+			if (pool_config.replication_enabled)
 			{
-				/* fatal error, notice to parent and exit */
-				notice_backend_error();
-				exit(1);
+				SECONDARY_CONNECTION(p) = malloc(sizeof(POOL_CONNECTION_POOL_SLOT));
+				SECONDARY_CONNECTION(p) = create_cp(SECONDARY_CONNECTION(p), user, database, 1);
+				p->num++;
 			}
 
-			p->user = strdup(user);
-			if (!p->user)
-			{
-				pool_error("pool_create_cp: out of memory");
-				return NULL;
-			}
-
-			p->database = strdup(database);
-			if (!p->database)
-			{
-				pool_error("pool_create_cp: out of memory");
-				return NULL;
-			}
-			p->con = pool_open(fd);
-			p->closetime = 0;
 			return p;
 		}
 		p++;
@@ -174,7 +170,7 @@ void pool_connection_pool_timer(POOL_CONNECTION_POOL *backend)
 
 	pool_debug("pool_connection_pool_timer: called");
 
-	backend->closetime = time(NULL);		/* set connection close time */
+	MASTER_CONNECTION(backend)->closetime = time(NULL);		/* set connection close time */
 
 	if (pool_config.connection_life_time == 0)
 		return;
@@ -182,10 +178,10 @@ void pool_connection_pool_timer(POOL_CONNECTION_POOL *backend)
 	/* look for any other timeout */
 	for (i=0;i<pool_config.max_pool;i++, p++)
 	{
-		if (p->user == NULL)
+		if (MASTER_CONNECTION(p)->user == NULL)
 			continue;
 
-		if (p != backend && p->closetime)
+		if (p != backend && MASTER_CONNECTION(p)->closetime)
 			return;
 	}
 
@@ -213,29 +209,29 @@ RETSIGTYPE pool_backend_timer_handler(int sig)
 
 	for (i=0;i<pool_config.max_pool;i++, p++)
 	{
-		if (p->user == NULL)
+		if (MASTER_CONNECTION(p)->user == NULL)
 			continue;
 
 		/* timer expire? */
-		if (p->closetime)
+		if (MASTER_CONNECTION(p)->closetime)
 		{
 			pool_debug("pool_backend_timer_handler: expire time: %d",
-					   p->closetime+pool_config.connection_life_time);
+					   MASTER_CONNECTION(p)->closetime+pool_config.connection_life_time);
 
-			if (now >= (p->closetime+pool_config.connection_life_time))
+			if (now >= (MASTER_CONNECTION(p)->closetime+pool_config.connection_life_time))
 			{
 				/* discard expired connection */
-				pool_debug("pool_backend_timer_handler: expires user %s database %s", p->user, p->database);
-				free(p->user);
-				free(p->database);
-				pool_close(p->con);
+				pool_debug("pool_backend_timer_handler: expires user %s database %s", MASTER_CONNECTION(p)->user, MASTER_CONNECTION(p)->database);
+				free(MASTER_CONNECTION(p)->user);
+				free(MASTER_CONNECTION(p)->database);
+				pool_close(MASTER_CONNECTION(p)->con);
 				memset(p, 0, sizeof(POOL_CONNECTION_POOL));
 			}
 			else
 			{
 				/* look for nearest timer */
-				if (p->closetime < nearest)
-					nearest = p->closetime;
+				if (MASTER_CONNECTION(p)->closetime < nearest)
+					nearest = MASTER_CONNECTION(p)->closetime;
 			}
 		}
 	}
@@ -251,13 +247,15 @@ RETSIGTYPE pool_backend_timer_handler(int sig)
 	}
 }
 
-int connect_inet_domain_socket(POOL_CONNECTION_POOL *p)
+int connect_inet_domain_socket(int secondary_backend)
 {
 	int fd;
 	int len;
 	int on = 1;
 	struct sockaddr_in addr;
 	struct hostent *hp;
+	int port;
+	char *host;
 
 	fd = socket(AF_INET, SOCK_STREAM, 0);
 	if (fd < 0)
@@ -279,13 +277,16 @@ int connect_inet_domain_socket(POOL_CONNECTION_POOL *p)
 	memset((char *) &addr, 0, sizeof(addr));
 	((struct sockaddr *)&addr)->sa_family = AF_INET;
 
-	addr.sin_port = htons(pool_config.current_backend_port);
+	host = secondary_backend?pool_config.secondary_backend_host_name:pool_config.current_backend_host_name;
+
+	port = secondary_backend?pool_config.secondary_backend_port:pool_config.current_backend_port;
+	addr.sin_port = htons(port);
 	len = sizeof(struct sockaddr_in);
 
-	hp = gethostbyname(pool_config.current_backend_host_name);
+	hp = gethostbyname(host);
 	if ((hp == NULL) || (hp->h_addrtype != AF_INET))
 	{
-		pool_error("connect_inet_domain_socket: gethostbyname() failed: %s",strerror(errno));
+		pool_error("connect_inet_domain_socket: gethostbyname() failed: %s host: %s", strerror(errno), host);
 		close(fd);
 		return -1;
 	}
@@ -302,11 +303,13 @@ int connect_inet_domain_socket(POOL_CONNECTION_POOL *p)
 	return fd;
 }
 
-int connect_unix_domain_socket(POOL_CONNECTION_POOL *p)
+int connect_unix_domain_socket(int secondary_backend)
 {
 	struct sockaddr_un addr;
 	int fd;
 	int len;
+	int port;
+	char *socket_dir;
 
 	fd = socket(AF_UNIX, SOCK_STREAM, 0);
 	if (fd == -1)
@@ -315,11 +318,11 @@ int connect_unix_domain_socket(POOL_CONNECTION_POOL *p)
 		return -1;
 	}
 
+	port = secondary_backend?pool_config.secondary_backend_port:pool_config.current_backend_port;
+	socket_dir = pool_config.backend_socket_dir;
 	memset((char *) &addr, 0, sizeof(addr));
 	((struct sockaddr *)&addr)->sa_family = AF_UNIX;
-	snprintf(addr.sun_path, sizeof(addr.sun_path), "%s/.s.PGSQL.%d", 
-			 pool_config.backend_socket_dir,
-			 pool_config.current_backend_port);
+	snprintf(addr.sun_path, sizeof(addr.sun_path), "%s/.s.PGSQL.%d", socket_dir, port);
 	pool_debug("connecting postmaster Unix domain socket: %s", addr.sun_path);
 
 	len = sizeof(struct sockaddr_un);
@@ -330,5 +333,41 @@ int connect_unix_domain_socket(POOL_CONNECTION_POOL *p)
 		close(fd);
 		return -1;
 	}
+	pool_debug("connected to postmaster Unix domain socket: %s fd: %d", addr.sun_path, fd);
 	return fd;
 }
+
+static POOL_CONNECTION_POOL_SLOT *create_cp(POOL_CONNECTION_POOL_SLOT *cp, char *user, char *database,
+									   int secondary_backend)
+{
+	int fd;
+
+	if (*pool_config.current_backend_host_name == '\0')
+		fd = connect_unix_domain_socket(secondary_backend);
+	else
+		fd = connect_inet_domain_socket(secondary_backend);
+
+	if (fd < 0)
+	{
+		/* fatal error, notice to parent and exit */
+		notice_backend_error();
+		exit(1);
+	}
+
+	cp->user = strdup(user);
+	if (!cp->user)
+	{
+		pool_error("pool_create_cp: out of memory");
+		return NULL;
+	}
+
+	cp->database = strdup(database);
+	if (!cp->database)
+	{
+		pool_error("pool_create_cp: out of memory");
+		return NULL;
+	}
+	cp->con = pool_open(fd);
+	cp->closetime = 0;
+	return cp;
+}
