@@ -1,6 +1,6 @@
 /* -*-pgsql-c-*- */
 /*
-* $Header: /home/t-ishii/repository/pgpool/pool_stream.c,v 1.11 2004/03/24 00:07:01 t-ishii Exp $
+* $Header: /home/t-ishii/repository/pgpool/pool_stream.c,v 1.15 2004/04/14 00:28:28 t-ishii Exp $
 *
 * pgpool: a language independent connection pool server for PostgreSQL 
 * written by Tatsuo Ishii
@@ -81,6 +81,8 @@ POOL_CONNECTION *pool_open(int fd)
 	cp->bufsz = READBUFSZ;
 	cp->po = 0;
 	cp->len = 0;
+	cp->sbuf = NULL;
+	cp->sbufsz = 0;
 
 	cp->fd = fd;
 	return cp;
@@ -94,6 +96,8 @@ void pool_close(POOL_CONNECTION *cp)
 	fclose(cp->read_fd);
 	fclose(cp->write_fd);
 	close(cp->fd);
+	free(cp->hp);
+	free(cp->sbuf);
 	free(cp);
 }
 
@@ -127,6 +131,15 @@ int pool_read(POOL_CONNECTION *cp, void *buf, int len)
 
 	while (len > 0)
 	{
+		if (cp->issecondary_backend)
+		{
+			if (pool_check_fd(cp, 0))
+			{
+				pool_log("pool_read: secondary data is not ready. abort this session");
+				exit(1);
+			}
+		}
+
 		readlen = read(cp->fd, buf, len);
 		if (readlen == -1)
 		{
@@ -155,7 +168,8 @@ int pool_read(POOL_CONNECTION *cp, void *buf, int len)
 			}
 			else
 			{
-				return -1;
+				/* protocol synchronization may be broken. better to exit this process... */
+				exit(1);
 			}
 		}
 		else if (len > readlen)
@@ -229,15 +243,14 @@ int pool_write_and_flush(POOL_CONNECTION *cp, void *buf, int len)
 */
 char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 {
-	static char *buf = NULL;
-	static int bufsz = 0;
-
 	int readp;
 	int readsize;
 	int readlen;
 	int strlength;
 	int flag;
 
+	int secondary = cp->issecondary_backend;
+
 #ifdef DEBUG
 	static char pbuf[READBUFSZ];
 #endif
@@ -246,16 +259,16 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 	readp = 0;
 
 	/* initialize read buffer */
-	if (bufsz == 0)
+	if (cp->sbufsz == 0)
 	{
-		buf = malloc(READBUFSZ);
-		if (buf == NULL)
+		cp->sbuf = malloc(READBUFSZ);
+		if (cp->sbuf == NULL)
 		{
 			pool_error("pool_read_string: malloc failed");
 			return NULL;
 		}
-		bufsz = READBUFSZ;
-		*buf = '\0';
+		cp->sbufsz = READBUFSZ;
+		*cp->sbuf = '\0';
 	}
 
 	/* any pending data? */
@@ -267,11 +280,11 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 			strlength = mystrlen(cp->hp+cp->po, cp->len, &flag);
 
 		/* buffer is too small? */
-		if ((strlength + 1) > bufsz)
+		if ((strlength + 1) > cp->sbufsz)
 		{
-			bufsz = ((strlength+1)/READBUFSZ+1)*READBUFSZ;
-			buf = realloc(buf, bufsz);
-			if (buf == NULL)
+			cp->sbufsz = ((strlength+1)/READBUFSZ+1)*READBUFSZ;
+			cp->sbuf = realloc(cp->sbuf, cp->sbufsz);
+			if (cp->sbuf == NULL)
 			{
 				pool_error("pool_read_string: realloc failed");
 				return NULL;
@@ -279,7 +292,7 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 		}
 
 		/* copy pending data to buffer */
-		memmove(buf, cp->hp+cp->po, strlength);
+		memmove(cp->sbuf, cp->hp+cp->po, strlength);
 
 #ifdef DEBUG
 		snprintf(pbuf, strlength, buf);
@@ -296,24 +309,24 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 			/* not null or line terminated */
 			cp->len = 0;
 			/* we need to read more since we have not encountered NULL yet */
-			readsize = bufsz - strlength;
+			readsize = cp->sbufsz - strlength;
 			readp = strlength;
 		}
 		else
 		{
-			return buf;
+			return cp->sbuf;
 		}
 	} else
 	{
-		readsize = bufsz;
+		readsize = cp->sbufsz;
 	}
 
 
 	for (;;)
 	{
-		pool_debug("pool_read_string: readsize: %d readp: %d", readsize, readp);
+		pool_debug("pool_read_string: readsize: %d readp: %d secondary: %d", readsize, readp, secondary);
 
-		readlen = read(cp->fd, buf+readp, readsize);
+		readlen = read(cp->fd, cp->sbuf+readp, readsize);
 		if (readlen == -1)
 		{
 			pool_error("pool_read_string: read() failed. reason:%s", strerror(errno));
@@ -329,13 +342,13 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 			}
 		}
 
-		pool_debug("pool_read_string: buf:%s: readlen: %d", buf, readlen);
+		pool_debug("pool_read_string: buf:%s: readlen: %d", cp->sbuf, readlen);
 
 		/* check overrun */
 		if (line)
-			strlength = mystrlinelen(buf+readp, readlen, &flag);
+			strlength = mystrlinelen(cp->sbuf+readp, readlen, &flag);
 		else
-			strlength = mystrlen(buf+readp, readlen, &flag);
+			strlength = mystrlen(cp->sbuf+readp, readlen, &flag);
 
 		pool_debug("strlength: %d", strlength);
 
@@ -347,11 +360,12 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 			/* pending buffer is enough? */
 			if (cp->len > cp->bufsz)
 			{
-				/* too small, enlarge it. we assume that pending data size is never larger than bufsz */
+				/* too small, enlarge it. we assume that pending data
+                   size is never larger than cp->sbufsz */
 
-				pool_debug("pool_read_string: enlarge pending buffer. old: %d new: %d", cp->bufsz, bufsz);
+				pool_debug("pool_read_string: enlarge pending buffer. old: %d new: %d", cp->bufsz, cp->sbufsz);
 
-				cp->hp = realloc(cp->hp, bufsz);
+				cp->hp = realloc(cp->hp, cp->sbufsz);
 				if (cp->hp == NULL)
 				{
 					pool_error("pool_read_string: realloc failed");
@@ -359,42 +373,42 @@ char *pool_read_string(POOL_CONNECTION *cp, int *len, int line)
 				}
 			}
 
-			memmove(cp->hp, buf+readp+strlength, cp->len);
+			memmove(cp->hp, cp->sbuf+readp+strlength, cp->len);
 			cp->po = 0;
 			*len += strlength;
-			pool_debug("pending data saved: :%s: length: %d return data: :%s: length: %d", cp->hp, cp->len, buf, *len);
+			pool_debug("pending data saved: :%s: length: %d return data: :%s: length: %d", cp->hp, cp->len, cp->sbuf, *len);
 
-			return buf;
+			return cp->sbuf;
 		}
 
 		*len += readlen;
 
-		pool_debug("pool_read_string: string: :%s:", buf+readp);
-		pool_debug("pool_read_string: next string: :%s:", buf+readp+strlen(buf+readp)+1);
+		pool_debug("pool_read_string: string: :%s:", cp->sbuf+readp);
+		pool_debug("pool_read_string: next string: :%s:", cp->sbuf+readp+strlen(cp->sbuf+readp)+1);
 		
 		/* enountered null or newline? */
 		if (flag)
 		{
 			/* ok we have read all data */
-			pool_debug("pool_read_string: final read %d string: :%s:", readlen, buf+readp);
-			pool_debug("pool_read_string: total result %d string: :%s:", *len, buf);
+			pool_debug("pool_read_string: final read %d string: :%s:", readlen, cp->sbuf+readp);
+			pool_debug("pool_read_string: total result %d string: :%s:", *len, cp->sbuf);
 			break;
 		}
 
-		pool_debug("pool_read_string: read more data %d string: :%s:", *len, buf+readp);
+		pool_debug("pool_read_string: read more data %d string: :%s:", *len, cp->sbuf+readp);
 
-		readp = bufsz;
-		bufsz += READBUFSZ;
+		readp = cp->sbufsz;
+		cp->sbufsz += READBUFSZ;
 		readsize = READBUFSZ;
 
-		buf = realloc(buf, bufsz);
-		if (buf == NULL)
+		cp->sbuf = realloc(cp->sbuf, cp->sbufsz);
+		if (cp->sbuf == NULL)
 		{
 			pool_error("pool_read_string: realloc failed");
 			return NULL;
 		}
 	}
-	return buf;
+	return cp->sbuf;
 }
 
 
